//!/////////////////////////////////////////////////////////////////////////
/// adc.c                                                                 //
///                                                                       //
/// Version:     0.3                                                      //
/// Date:        11/19/2009                                               //
/// Author:      Michael Skutek (michael@skutek.de)                       //
///                                                                       //
/// Description: A/D conversion function (single conversion)              //
///                                                                       //
/// Platform:    AVR ATmega (tested on ATMega16)                          //
///                                                                       //
/// Include Files:                                                        //
///              avr/io.h:             port accessing (part of avr-gcc)   //
///              port_definitions.h:   definition of output ports         //
///              adc.h:                defintions (e.g. return values)    //
///                                                                       //
///                                                                       //
/// This program is free software: you can redistribute it and/or modify  //
/// it under the terms of the GNU Lesser General Public License as        //
/// published by the Free Software Foundation, either version 3 of the    //
/// License, or (at your option) any later version.                       //
///                                                                       //
/// This program is distributed in the hope that it will be useful,       //
/// but WITHOUT ANY WARRANTY; without even the implied warranty of        //
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          //
/// GNU Lesser General Public License for more details.                   //
///                                                                       //
/// You should have received a copy of the GNU Lesser General Public      //
/// License along with this program. If not, see                          //
/// <http://www.gnu.org/licenses/>.                                       //
//!/////////////////////////////////////////////////////////////////////////




#include <stdint.h>
#include <avr/io.h>

#include "port_definitions.h"
#include "adc.h"




//!///////////////////////////////////////////////////////////////////////
///single_adc_read                                                      //
///                                                                     //
/// Description:                                                        //
///         Single read of one adc pin.                                 //
///                                                                     //
/// input:                                                              //
///      v_ref_reg:     reference voltage (content of REFS0 and REFS1)  //
///      v_ref_val:     reference voltage [mV]                          //
///  number_cycles:     number of cycles for conversion (mean value)    //
///       min_freq:     minimum frequency for conversion [Hz]           //
///                                                                     //
/// return:                                                             //
///         voltage [mV]                                                //
///                                                                     //
//!///////////////////////////////////////////////////////////////////////

uint32_t single_adc_read(uint8_t vref_reg, uint16_t vref_val,
                         uint16_t number_cycles, uint16_t min_freq) {

  uint32_t ad_voltage;
  uint16_t tmp_voltage;
  uint16_t frequency_prescaler;
  uint16_t i;


  ///some initialization
  ad_voltage = ADC_UNKNOWN_ERROR;
  ADCSRA = 0x00;                        //ADCSRA register = 0x00  --> ADC disabled completely


  ///check input values

  //A/D pin needs the following configuration: input pin, internal pull-up resistor deactivated
  if ((((ADC_DDR >> ADC_PIN) & 1) != 0) || (((ADC_PORT >> ADC_PIN) & 1) != 0)) {
    return ADC_PIN_ERROR;
  }

  //check number of requested conversion cycles
  if ((number_cycles > MAX_NUMBER_ADC_CYCLES) || (number_cycles < 1)) {
    return ADC_NUMBERCYCLES_ERROR;
  }

  //check given ADC frequency
  if ((min_freq > MAX_ADC_FREQUENCY) || (min_freq < MIN_ADC_FREQUENCY)) {
    return ADC_FREQUENCY_ERROR;
  }


  ///set ADC register values

  //set ADMUX register
  ADMUX = ADC_PIN | (vref_reg << REFS0);

  //calculate the value for frequency setting
  frequency_prescaler = F_CPU / (min_freq * 1000);

  //set AD Prescaler Bits depending on Prescaler value and minimum frequency and start the first conversion
  if (frequency_prescaler > 64) {
    ADCSRA = 0x00 | (1 << ADEN) | (1 << ADSC) | (1 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); //prescaler=128
  } else if (frequency_prescaler > 32) {
    ADCSRA = 0x00 | (1 << ADEN) | (1 << ADSC) | (1 << ADPS2) | (1 << ADPS1) | (0 << ADPS0); //prescaler=64
  } else if (frequency_prescaler > 16) {
    ADCSRA = 0x00 | (1 << ADEN) | (1 << ADSC) | (1 << ADPS2) | (0 << ADPS1) | (1 << ADPS0); //prescaler=32
  } else if (frequency_prescaler > 8) {
    ADCSRA = 0x00 | (1 << ADEN) | (1 << ADSC) | (1 << ADPS2) | (0 << ADPS1) | (0 << ADPS0); //prescaler=16
  } else if (frequency_prescaler > 4) {
    ADCSRA = 0x00 | (1 << ADEN) | (1 << ADSC) | (0 << ADPS2) | (1 << ADPS1) | (1 << ADPS0); //prescaler=8
  } else if (frequency_prescaler > 2) {
    ADCSRA = 0x00 | (1 << ADEN) | (1 << ADSC) | (0 << ADPS2) | (1 << ADPS1) | (0 << ADPS0); //prescaler=4
  } else {
    ADCSRA = 0x00 | (1 << ADEN) | (1 << ADSC) | (0 << ADPS2) | (0 << ADPS1) | (1 << ADPS0); //prescaler=2
  }

  //wait until the first conversion has been completed
  while (((ADCSRA >> ADSC) & 1) != 0) {
    ;    //do nothing
  }

  //now the first conversion has been completed --> read the result
  tmp_voltage = ADCL;
  tmp_voltage = ad_voltage | (ADCH << 8);

  ad_voltage = 0;

  //now start <number_cycles> conversions and calculate the mean
  for (i = 0; i < number_cycles; i++) {

    ADCSRA = ADCSRA | (1 << ADSC);  //start conversion

    //wait until the first conversion has been completed
    while (((ADCSRA >> ADSC) & 1) != 0) {
      ;    //do nothing
    }

    //now the first conversion has been completed --> read the result
    tmp_voltage = ADCL;
    tmp_voltage = tmp_voltage | (ADCH << 8);

    ad_voltage = ad_voltage + tmp_voltage;

  }


  //disable the ADC
  ADCSRA = 0x00;


  //calculate the result (mean value)
  ad_voltage = ad_voltage * vref_val;
  ad_voltage = ad_voltage / number_cycles;
  ad_voltage = ad_voltage >> 10;          //ad_voltage/1024

  return ad_voltage;

}



//!///////////////////////////////////////////////////////////////////////
///adc_pin_init                                                         //
///                                                                     //
/// Description:                                                        //
///         Initializes the ATmega I/O pin for A/D conversion.          //
///                                                                     //
/// input:                                                              //
///                -                                                    //
///                                                                     //
/// return:                                                             //
///                -                                                    //
///                                                                     //
//!///////////////////////////////////////////////////////////////////////


void adc_pin_init(void) {


  //set ADC Port Pin to input pin (no pull-up resistor)
  ADC_DDR = ADC_DDR & ~(1 << ADC_PIN);
  ADC_PORT = ADC_PORT & ~(1 << ADC_PIN);


}


