//!/////////////////////////////////////////////////////////////////////////
/// isr.c                                                                //
///                                                                       //
/// Version:     0.3                                                      //
/// Date:        05/22/2010                                               //
/// Author:      Michael Skutek (michael@skutek.de)                       //
///                                                                       //
/// Description: Contains the interrupt service routines                  //
///                                                                       //
/// Platform:    AVR ATmega (tested on ATmega16)                          //
///                                                                       //
///                                                                       //
///                                                                       //
/// This program is free software: you can redistribute it and/or modify  //
/// it under the terms of the GNU Lesser General Public License as        //
/// published by the Free Software Foundation, either version 3 of the    //
/// License, or (at your option) any later version.                       //
///                                                                       //
/// This program is distributed in the hope that it will be useful,       //
/// but WITHOUT ANY WARRANTY; without even the implied warranty of        //
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          //
/// GNU Lesser General Public License for more details.                   //
///                                                                       //
/// You should have received a copy of the GNU Lesser General Public      //
/// License along with this program. If not, see                          //
/// <http://www.gnu.org/licenses/>.                                       //
//!/////////////////////////////////////////////////////////////////////////


#include <stdint.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

#include "babyfone_receiver.h"
#include "port_definitions.h"
#include "usart.h"
#include "rfm12_io.h"


///global variables for ISR
extern volatile char usart_rcv_buffer[USART_RCV_BUFFER_LENGTH];
extern volatile uint8_t usart_rcv_position;
extern volatile uint8_t print_command_error;
extern volatile uint8_t print_command_info;

extern volatile uint16_t second_completed_counter;
extern volatile uint8_t second_completed_flag;

extern volatile uint16_t receive_buffer[RFM12_RXBUFFER_LENGTH];
extern volatile uint16_t tx_data;
extern volatile uint32_t checksum;
extern volatile uint8_t receive_counter;
extern volatile uint8_t packet_received;



//!/////////////////////////////////////////////
///Timer ISR for buttons////////////////////////
//!/////////////////////////////////////////////

#if MCU == 644
ISR(TIMER0_COMPA_vect) {
#else
ISR(TIMER0_COMP_vect) {
#endif

  if (second_completed_counter < SECOND_COMPLETED_MAX) {
    second_completed_counter++;
  } else {
    second_completed_flag = TRUE;
    second_completed_counter = 0;
  }

}



//!/////////////////////////////////////////////
///USART ISR Receiving//////////////////////////
//!/////////////////////////////////////////////

#if MCU == 644
ISR(USART0_RX_vect) {
#else
ISR(USART_RXC_vect) {
#endif

  uint8_t return_value;

#if MCU == 644
  usart_rcv_buffer[usart_rcv_position] = UDR0;
#else
  usart_rcv_buffer[usart_rcv_position] = UDR;
#endif

  if (usart_rcv_buffer[usart_rcv_position] == '+') {

    return_value = process_usart_command();   //process received command

    //check for return value --> command ok or unknown?
    if (return_value == USART_COMMAND_OK) {
      usart_rcv_position = 0;           //reset buffer position
    } else {
      print_command_error = UART_UNKNOWN_COMMAND_MSG;
      usart_rcv_position = 0;           //reset buffer position
    }

  } else if (usart_rcv_buffer[usart_rcv_position] == '*') {

    print_command_info = UART_RCV_BUFFER_RESET_MSG;
    //reset buffer position
    usart_rcv_position = 0;

  } else if (usart_rcv_buffer[usart_rcv_position] == '\n') {

    usart_rcv_position = 0;

  } else if (usart_rcv_position < (USART_RCV_BUFFER_LENGTH - 1)) {

    usart_rcv_position++;

  } else {

    print_command_error = UART_RCV_BUFFER_OVERFLOW_MSG;
    //reset buffer position
    usart_rcv_position = 0;

  }

}




//!/////////////////////////////////////////////
///EXTERNAL ISR Receiving (RFM12 IRQ)///////////
//!/////////////////////////////////////////////

ISR(INT2_vect) {

  uint16_t dummy;


  ///read status register
  tx_data = RFM12_STATUSREAD_VALUE;
  rfm12_data_io(&tx_data, &receive_buffer[receive_counter], 1, RFM12_REGISTER_DATA);


  ///read RX FIFO if FFIT Bit in Status Register is set
  if ((receive_buffer[receive_counter] & 0x8000) > 0) {

    //read FIFO Buffer
    tx_data = RFM12_FIFOREAD_VALUE;
    rfm12_data_io(&tx_data, &receive_buffer[receive_counter], 1, RFM12_RX_DATA);

    //if complete packet received - disable and re-enable FIFO
    //FIFO will then wait for the next sync pattern
    if (receive_counter < (RFM12_RXPACKET_LENGTH - 1)) {
      receive_counter++;
    } else {
      tx_data = RFM12_NOFIFOMODE_VALUE;
      rfm12_data_io(&tx_data, &dummy, 1, RFM12_REGISTER_DATA);
      tx_data = RFM12_FIFOMODE_VALUE;
      rfm12_data_io(&tx_data, &dummy, 1, RFM12_REGISTER_DATA);
      receive_counter = 0;
      packet_received = TRUE;
    }

  }


}
