//!/////////////////////////////////////////////////////////////////////////
/// main.c                                                                //
///                                                                       //
/// Version:     0.5                                                      //
/// Date:        03/04/2013                                               //
/// Author:      Michael Skutek (michael@skutek.de)                       //
///                                                                       //
/// Description: Main loop of Babyphone Receiver software                 //
///                                                                       //
/// Platform:    AVR ATmega (tested on ATmega16)                          //
///                                                                       //
///                                                                       //
///                                                                       //
/// This program is free software: you can redistribute it and/or modify  //
/// it under the terms of the GNU Lesser General Public License as        //
/// published by the Free Software Foundation, either version 3 of the    //
/// License, or (at your option) any later version.                       //
///                                                                       //
/// This program is distributed in the hope that it will be useful,       //
/// but WITHOUT ANY WARRANTY; without even the implied warranty of        //
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          //
/// GNU Lesser General Public License for more details.                   //
///                                                                       //
/// You should have received a copy of the GNU Lesser General Public      //
/// License along with this program. If not, see                          //
/// <http://www.gnu.org/licenses/>.                                       //
//!/////////////////////////////////////////////////////////////////////////


#include <avr/io.h>
#include <stdint.h>
#include <stdlib.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "babyfone_receiver.h"
#include "port_definitions.h"
#include "lcd.h"
#include "glcd.h"
#include "usart.h"
#include "rfm12_io.h"


///some definitions

//Definition for Frequency Setting
#ifndef F_CPU
#warning "F_CPU was not defined! It will now be set to 16000000UL."
#define F_CPU 16000000UL    // Clock Cycle [Hz]
//  #warning "F_CPU was not defined! It will now be set to 7372800UL."
//  #define F_CPU 7372800UL    // Clock Cycle [Hz]
#endif

//definition for baud rate (RS232)
#define BAUD 38400UL          // baud rate for RS232

#define UBRR_VAL ((F_CPU+BAUD*8)/(BAUD*16)-1)   //Value for UBRR
#define BAUD_REAL (F_CPU/(16*(UBRR_VAL+1)))     //real baud rate
#define BAUD_ERROR ((BAUD_REAL*1000)/BAUD)      // Error [1/oo], 1000 = no error.

//#if ((BAUD_ERROR<990) || (BAUD_ERROR>1010))     //check for systematic error
//#error Systematic Error of baud rate more than 1%!
//#endif


///global variables for ISR
volatile uint8_t prev_BUT_PIN = 0xff;

volatile char usart_rcv_buffer[USART_RCV_BUFFER_LENGTH];
volatile uint8_t usart_rcv_position;

volatile uint8_t print_command_error;
volatile uint8_t print_command_info;

volatile uint16_t second_completed_counter;
volatile uint8_t second_completed_flag;


volatile uint16_t receive_buffer[RFM12_RXBUFFER_LENGTH];
volatile uint16_t tx_data;
volatile uint32_t checksum;
volatile uint8_t receive_counter;
volatile uint8_t packet_received;



///static string messages --> FLASH memory
const char startup_MSG[] PROGMEM = "Babyfon Receiver Version 0.5";
const char startup_MSG02[] PROGMEM = "Baby ruhig...";
const char clear_line_MSG[] PROGMEM = "                ";
const char time_reset_MSG[] PROGMEM = "Zeit: 00:00:00 *";
const char time_MSG[] PROGMEM = "Zeit: ";
const char micro_MSG[] PROGMEM = "Mikro";
const char mute_MSG[] PROGMEM = "   **STUMM**   ";
const char no_connection_MSG[] PROGMEM = "keine Verbindung";
const char rfm_link_err_MSG[] PROGMEM = "RFM Link ERR";
const char baby_detect_info_MSG[] PROGMEM = "\nINFO: Acoustic Signal [mV]: ";
const char error_info_MSG01[] PROGMEM = "USART: ERROR: Unknown Command received!\n";
const char error_info_MSG02[] PROGMEM = "Valid commands are:\n";
const char error_info_MSG03[] PROGMEM = "  synchronize+                 --> synchronizes to receiver\n";
const char error_info_MSG10[] PROGMEM = "USART: ERROR: Receive Buffer Overflow! Please reset buffer with \'*\'!\n";
const char alive_MSG[] PROGMEM = "INFO: Babyfone is running.\n";




//!/////////////////////////////////////////////
///main function////////////////////////////////
//!/////////////////////////////////////////////

int main(void) {

  uint8_t i;
  uint32_t checksum_test;
  uint32_t detection_buffer_value;
  char tmp_char[35];

  uint8_t seconds, minutes;
  uint16_t hours;

  uint8_t backlight_counter;

  uint16_t mute_counter;

  uint8_t button_counter;

  uint32_t sync_counter;

  uint8_t first_detection;

  uint8_t baby_detected_flag;




  ///port/pin configuration

  //setup LCD I/O pins
  lcd_pin_init();

  LCD_PORT = LCD_PORT | (1 << LCD_BACKLIGHT);
  LCD_DDR = LCD_DDR | (1 << LCD_BACKLIGHT);


  //setup for RFM12 Wireless Module
  rfm12_pin_init();


  //setup buzzer, button and LED
  BUZ_PORT = BUZ_PORT & ~(1 << BUZ_ADDR0);   //init value: 0
  BUZ_DDR = BUZ_DDR | (1 << BUZ_ADDR0);      //output port

  BUT_PORT = BUT_PORT | (1 << BUT_ADDR);     //enable pull-up resistor(s)
  BUT_DDR = BUT_DDR & ~(1 << BUT_ADDR);      //input port

  LED_PORT = LED_PORT & ~(1 << LED_ADDR0);   //init value: 0
  LED_DDR = LED_DDR | (1 << LED_ADDR0);      //output port




  ///USART configuration
  //enable UART for sending something
  USART_Init(UBRR_VAL);
  usart_rcv_position = 0;
  print_command_error = UART_NO_ERROR_MSG;
  print_command_info = UART_NO_INFO_MSG;


  ///set up 8 Bit Timer ISR
#if MCU == 644
  //TCCR0A = 0x05;          //clock cycle/1024
  TCCR0A = (1 << WGM01);    //CTC is set, pin is not connected
  TCCR0B = (1 << CS02) | (1 << CS00);     //Clock/1024
  OCR0A = SECOND_CTR_VALUE;                //value for compare match
  TIMSK0 = TIMSK0 | (1 << OCIE0A); //enable interrupt
#else
  TCCR0 = 0x00 | (1 << WGM01) | (1 << CS02) | (1 << CS00);     //clock cycle/1024, CTC mode
  OCR0 = SECOND_CTR_VALUE;                                     //value for compare match
  TIMSK = TIMSK | (1 << OCIE0); //enable interrupt
#endif


  ///initialize counter and flags
  second_completed_counter = 0;
  second_completed_flag = FALSE;
  seconds = 0;
  minutes = 0;
  hours = 0;
  first_detection = FALSE;

  baby_detected_flag = FALSE;

  sync_counter = MAX_SYNC_TIME;

  backlight_counter = 0;

  mute_counter = 0;

  button_counter = 0;


  ///enable external interrupt for RFM12
  GICR = GICR & ~(1 << INT2);               //disable INT2
  MCUCSR = MCUCSR & ~(1 << ISC2);           //Interrupt will be activated on falling edge
  GICR = GICR | (1 << INT2);                //enable INT2


  ///initialize LCD
  lcd_init();


  ///initialize the RFM12 Wireless Module
  rfm12_init();


  ///activate global Interrupts
  sei();


  ///initialize receive buffer
  receive_counter = 0;
  packet_received = FALSE;
  detection_buffer_value = 0;
  for (i = 0; i < RFM12_RXBUFFER_LENGTH; i++) {
    receive_buffer[i] = 0;
  }


  ///startup message
  lcd_clear();
  lcd_cursor_home();
  lcd_send_string(startup_MSG, TRUE);

  _delay_ms(2000);

  /*lcd_clear();
  lcd_send_string(startup_MSG02, TRUE);
  lcd_set_cursor(0, 1);*/



  ///endless loop (main function)
  while (1) {


    ///reset baby detected flag...
    baby_detected_flag = FALSE;

    ///reset BUZZER and LED Output
    BUZ_PORT = BUZ_PORT & ~(1 << BUZ_ADDR0);
    LED_PORT = LED_PORT & ~(1 << LED_ADDR0);

    ///get Button state
    if ((BUT_PIN & (1 << BUT_ADDR)) == 0) {      //if button pressed - switch on backlight
      backlight_counter = BACKLIGHT_ON_TIME1;
      LCD_PORT = LCD_PORT | (1 << LCD_BACKLIGHT);
    } else {
      button_counter = 0;                       //reset button counter  --> button counter will be set every second completed
    }


    ///if a complete packet has been received...
    if (packet_received == TRUE) {

      packet_received = FALSE;
      checksum_test = 0;
      checksum = 0;
      detection_buffer_value = 0;

      for (i = 0; i < RFM12_RXPACKET_LENGTH; i++) {

        //extract voltage data
        if (i == 3) {
          detection_buffer_value = detection_buffer_value | (receive_buffer[i] & 0x00ff);
        } else if (i == 2) {
          detection_buffer_value = detection_buffer_value | (((uint32_t) (receive_buffer[i] & 0x00ff) << 8) & 0x0000ff00);
        } else if (i == 1) {
          detection_buffer_value = detection_buffer_value | (((uint32_t) (receive_buffer[i] & 0x00ff) << 16) & 0x00ff0000);
        } else if (i == 0) {
          detection_buffer_value = detection_buffer_value | (((uint32_t) (receive_buffer[i] & 0x00ff) << 24) & 0xff000000);
        }

        //extract checksum from data packet
        if (i == RFM12_CHECKSUM_BYTE0_ID) {
          checksum = checksum | (receive_buffer[i] & 0x00ff);
        } else if (i == RFM12_CHECKSUM_BYTE1_ID) {
          checksum = checksum | (((uint32_t) (receive_buffer[i] & 0x00ff) << 8) & 0x0000ff00);
        } else if (i == RFM12_CHECKSUM_BYTE2_ID) {
          checksum = checksum | (((uint32_t) (receive_buffer[i] & 0x00ff) << 16) & 0x00ff0000);
        } else if (i == RFM12_CHECKSUM_BYTE3_ID) {
          checksum = checksum | (((uint32_t) (receive_buffer[i] & 0x00ff) << 24) & 0xff000000);
        }


        //calculate the checksum of received data
        if (i < RFM12_DATA_LENGTH) {
          checksum_test = checksum_test + (receive_buffer[i] & 0x00ff);
        }

      }

      ///print received value in second LCD line
      lcd_set_cursor(0, 1);
      lcd_send_string(clear_line_MSG, TRUE);

      //calculate mic voltage value
      detection_buffer_value = detection_buffer_value / DETECTION_BUFFER_LENGTH;

      //check checksum...
      if (checksum == checksum_test) {

        lcd_set_cursor(0, 1);
        lcd_send_string(micro_MSG, TRUE);
        itoa((int32_t)((receive_buffer[RFM12_BABYDETECT_BYTE_ID] >> 4) & 0x0f), tmp_char, 10);
        lcd_send_string(tmp_char, FALSE);
        lcd_send_string(": ", FALSE);
        itoa((int32_t)detection_buffer_value, tmp_char, 10);
        lcd_send_string(tmp_char, FALSE);
        lcd_send_string(" mV", FALSE);



        ///reset time, activate buzzer and backlight if baby was detected
        if ((receive_buffer[RFM12_BABYDETECT_BYTE_ID] & 0x0f) > 0) {            //first byte: detection indication, second byte: sender ID

          baby_detected_flag = TRUE;
          backlight_counter = BACKLIGHT_ON_TIME2;
          LCD_PORT = LCD_PORT | (1 << LCD_BACKLIGHT);

          //switch on buzzer only if no mute time is running
          if ((mute_counter == 0) && (detection_buffer_value > DETECTION_BUFFER_TRESHOLD)) {
            BUZ_PORT = BUZ_PORT | (1 << BUZ_ADDR0);
          }

          LED_PORT = LED_PORT | (1 << LED_ADDR0);

          hours = 0;
          minutes = 0;
          seconds = 0;
          first_detection = TRUE;       //reset startup flag
          print_command_info = UART_BABY_DETECT_MSG;  //send UART message

          //update display
          lcd_set_cursor(0, 0);
          lcd_send_string(clear_line_MSG, TRUE);
          lcd_set_cursor(0, 0);
          lcd_send_string(time_reset_MSG, TRUE);

        } else {

          print_command_info = UART_ALIVE_MSG;        //send UART message

          //switch off backlight (if not requested by user)
          if (backlight_counter == 0) {
            LCD_PORT = LCD_PORT & ~(1 << LCD_BACKLIGHT);
          }

        }

        sync_counter = 0;                 //reset sync counter

      } else {
        lcd_set_cursor(0, 1);
        lcd_send_string(rfm_link_err_MSG, TRUE);
        print_command_info = UART_LINK_ERROR_MSG;        //send UART message
      }

    }     //if (packet_received == TRUE) {...




    ///update internal timer
    if (second_completed_flag == TRUE) {    //a new second was completed  --> will be true for each second

      if (seconds < 59) {

        seconds++;

      } else {

        if (minutes < 59) {
          minutes++;
        } else {
          hours++;
          minutes = 0;
        }

        seconds = 0;

      }

      second_completed_flag = FALSE;

      if (first_detection == TRUE) {            //baby was detected some time before --> print elapsed time

        lcd_set_cursor(0, 0);
        lcd_send_string(clear_line_MSG, TRUE);
        lcd_set_cursor(0, 0);
        lcd_send_string(time_MSG, TRUE);

        if (hours < 10) lcd_send_string("0", FALSE);
        itoa(hours, tmp_char, 10);
        lcd_send_string(tmp_char, FALSE);
        lcd_send_string(":", FALSE);

        if (minutes < 10) lcd_send_string("0", FALSE);
        itoa(minutes, tmp_char, 10);
        lcd_send_string(tmp_char, FALSE);
        lcd_send_string(":", FALSE);

        if (seconds < 10) lcd_send_string("0", FALSE);
        itoa(seconds, tmp_char, 10);
        lcd_send_string(tmp_char, FALSE);

      }

      ///do some things synchronized to second timer

      //decrease backlight counter if > 0
      if (backlight_counter > 0) backlight_counter--;

      //check and set sync counter
      if (sync_counter > MAX_SYNC_TIME) {
        if ((seconds & 0x1) > 0) {
          lcd_set_cursor(0, 1);
          lcd_send_string(clear_line_MSG, TRUE);
          LCD_PORT = LCD_PORT & ~(1 << LCD_BACKLIGHT);
        } else {
          lcd_set_cursor(0, 1);
          lcd_send_string(no_connection_MSG, TRUE);
          LCD_PORT = LCD_PORT | (1 << LCD_BACKLIGHT);
        }
      } else {

        //switch off backlight after specified time
        if (backlight_counter == 0) {
          LCD_PORT = LCD_PORT & ~(1 << LCD_BACKLIGHT);
        }


      }

      //decrease mute counter if > 0
      if (mute_counter > 0) {
        mute_counter--;
        //create blinking mute flag (with each second)
        if ((seconds & 0x1) > 0) {
          lcd_set_cursor(0, 1);
          lcd_send_string(clear_line_MSG, TRUE);
          lcd_set_cursor(0, 1);
          lcd_send_string(mute_MSG, TRUE);       //print mute message
        } else {
          if (sync_counter <= MAX_SYNC_TIME) {
            lcd_set_cursor(0, 1);
            lcd_send_string(clear_line_MSG, TRUE);
            lcd_set_cursor(0, 1);
            lcd_send_string(micro_MSG, TRUE);
            itoa((int32_t)((receive_buffer[RFM12_BABYDETECT_BYTE_ID] >> 4) & 0x0f), tmp_char, 10);
            lcd_send_string(tmp_char, FALSE);
            lcd_send_string(": ", FALSE);
            itoa((int32_t)detection_buffer_value, tmp_char, 10);
            lcd_send_string(tmp_char, FALSE);
            lcd_send_string(" mV", FALSE);
          }
        }
        //clear mute flag if mute time exceeded
        if (mute_counter == 0) {
          if (sync_counter <= MAX_SYNC_TIME) {
            lcd_set_cursor(0, 1);
            lcd_send_string(clear_line_MSG, TRUE);
            lcd_set_cursor(0, 1);
            lcd_send_string(micro_MSG, TRUE);
            itoa((int32_t)((receive_buffer[RFM12_BABYDETECT_BYTE_ID] >> 4) & 0x0f), tmp_char, 10);
            lcd_send_string(tmp_char, FALSE);
            lcd_send_string(": ", FALSE);
            itoa((int32_t)detection_buffer_value, tmp_char, 10);
            lcd_send_string(tmp_char, FALSE);
            lcd_send_string(" mV", FALSE);
          } else {
            lcd_set_cursor(0, 1);
            lcd_send_string(clear_line_MSG, TRUE);
          }
        }
      }

      //check for button pressing for a longer time to activate muting
      if ((BUT_PIN & (1 << BUT_ADDR)) == 0) {
        if (button_counter < BUTTON_MUTE_TIME) {
          button_counter++;                         //increrase mute counter (will be set to 0 if button was released)
        } else {
          //buzzer mute
          mute_counter = BUZZER_MUTE_TIME;
          lcd_set_cursor(0, 1);
          lcd_send_string(clear_line_MSG, TRUE);
          lcd_set_cursor(0, 1);
          lcd_send_string(mute_MSG, TRUE);       //print mute message
        }
      }


      sync_counter++;

    }



    ///check for error message from ISR and send USART messages
    if (print_command_error == UART_UNKNOWN_COMMAND_MSG) {
      USART_puts(error_info_MSG01, TRUE);                                 //put string to serial port
      USART_puts(error_info_MSG02, TRUE);                                 //put string to serial port
      USART_puts(error_info_MSG03, TRUE);                                 //put string to serial port
      print_command_error = UART_NO_ERROR_MSG;
    } else if (print_command_error == UART_RCV_BUFFER_OVERFLOW_MSG) {
      USART_puts(error_info_MSG10, TRUE);                //put string to serial port
      print_command_error = UART_NO_ERROR_MSG;
    }

    ///check for pending info message from ISR or other functions and send USART messages
    if (print_command_info == UART_BABY_DETECT_MSG) {
      USART_puts(baby_detect_info_MSG, TRUE);                //put string to serial port
      itoa(detection_buffer_value, tmp_char, 10);
      USART_puts(tmp_char, FALSE);
      print_command_info = UART_NO_INFO_MSG;
    } else if (print_command_info == UART_ALIVE_MSG) {
      USART_puts(alive_MSG, TRUE);                //put string to serial port
      print_command_info = UART_NO_INFO_MSG;
    }  else if (print_command_info == UART_LINK_ERROR_MSG) {
      USART_puts(rfm_link_err_MSG, TRUE);                //put string to serial port
      print_command_info = UART_NO_INFO_MSG;
    }



  }

}





