//!/////////////////////////////////////////////////////////////////////////
/// rfm12_io.c                                                            //
///                                                                       //
/// Version:     0.1                                                      //
/// Date:        11/28/2009                                               //
/// Author:      Michael Skutek (michael@skutek.de)                       //
///                                                                       //
/// Description: SPI Interface to RFM2 wireless module                    //
///                                                                       //
/// Platform:    AVR ATmega (tested on ATmega16)                          //
///                                                                       //
/// Include Files:                                                        //
///              avr/io.h:             port accessing (part of avr-gcc)   //
///              util/delay.h:         time delays (part of avr-gcc)      //
///              port_definitions.h:   definition of output ports         //
///              rfm12_io.h:           definitions for RFM12_io           //
///                                                                       //
///                                                                       //
/// This program is free software: you can redistribute it and/or modify  //
/// it under the terms of the GNU Lesser General Public License as        //
/// published by the Free Software Foundation, either version 3 of the    //
/// License, or (at your option) any later version.                       //
///                                                                       //
/// This program is distributed in the hope that it will be useful,       //
/// but WITHOUT ANY WARRANTY; without even the implied warranty of        //
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          //
/// GNU Lesser General Public License for more details.                   //
///                                                                       //
/// You should have received a copy of the GNU Lesser General Public      //
/// License along with this program. If not, see                          //
/// <http://www.gnu.org/licenses/>.                                       //
//!/////////////////////////////////////////////////////////////////////////

#include <avr/io.h>
#include <util/delay.h>
#include <stdint.h>
#include <stdlib.h>

#include "port_definitions.h"
#include "rfm12_io.h"
#include "glcd.h"



//!///////////////////////////////////////////////////////////////////////
///rfm12_write_data                                                     //
///                                                                     //
/// Description:                                                        //
///         Sends 2 bytes to RFM12 Wireless Module (SPI).               //
///                                                                     //
/// input:                                                              //
///           tx_data:     data words to send (MOSI)                    //
///           rx_data:     received data words (MISO)                   //
///      number_bytes:     number of bytes to receive/send              //
///           di_flag:     mode: rx data, tx data, register i/o         //
///                                                                     //
/// return:                                                             //
///        byte_count:     number of read or sent bytes                 //
///                                                                     //
//!///////////////////////////////////////////////////////////////////////

uint8_t rfm12_data_io(volatile uint16_t *tx_data, volatile uint16_t *rx_data, uint8_t number_bytes, uint8_t di_flag) {

  uint8_t i;
  uint8_t byte_count = 0;


  ///read/write register
  if (di_flag == RFM12_REGISTER_DATA) {

    //Start of transmission --> Chip Select to '0'
    RFM12_PORT = RFM12_PORT & ~(1 << RFM12_nSEL);

    SPDR = (*tx_data >> 8) & 0xff;         //send high byte to slave

    //wait until data transmission to RFM12 is complete...
    while (!(SPSR & (1 << SPIF))) {
      ;
    }
    *rx_data = SPDR;     //read high byte from slave

    SPDR = *tx_data & 0xff;         //send low byte to slave

    //wait until data transmission to RFM12 is complete
    while (!(SPSR & (1 << SPIF))) {
      ;
    }
    *rx_data = (*rx_data << 8) | SPDR;    //read low byte from slave

    //end of transmission --> Chip Select to '1'
    RFM12_PORT = RFM12_PORT | (1 << RFM12_nSEL);

    byte_count++;



    ///write into TX Register
  } else if (di_flag == RFM12_TX_DATA) {


    //TX LED on
    RFM12_PORT = RFM12_PORT | (1 << RFM12_LEDY);

    //Start of transmission --> Chip Select to '0'
    RFM12_PORT = RFM12_PORT & ~(1 << RFM12_nSEL);


    ///send Config Command Message --> enable TX register "EL"
    SPDR = (RFM12_CONFIG_VALUE >> 8) & 0xff;         //send high byte to slave

    //wait until data transmission to RFM12 is complete...
    while (!(SPSR & (1 << SPIF))) {
      ;
    }
    SPDR = RFM12_CONFIG_VALUE & 0xff;         //send low byte to slave

    //wait until data transmission to RFM12 is complete
    while (!(SPSR & (1 << SPIF))) {
      ;
    }

#if RFM12_TRANSMISSION_BURST == 0
    //end of transmission --> Chip Select to '1'
    RFM12_PORT = RFM12_PORT | (1 << RFM12_nSEL);

    _delay_us(0.01);

    //Start of transmission --> Chip Select to '0'
    RFM12_PORT = RFM12_PORT & ~(1 << RFM12_nSEL);
#endif

    ///send Power Management Config Message --> enable TX register "ET" (switch on transmitter)
    SPDR = (RFM12_POWERMANAGEMENT_TX_VALUE >> 8) & 0xff;         //send high byte to slave

    //wait until data transmission to RFM12 is complete...
    while (!(SPSR & (1 << SPIF))) {
      ;
    }
    SPDR = RFM12_POWERMANAGEMENT_TX_VALUE & 0xff;         //send low byte to slave

    //wait until data transmission to RFM12 is complete
    while (!(SPSR & (1 << SPIF))) {
      ;
    }

    //wait until RFM12 is ready for the next byte
    while (!(RFM12_PIN & (1 << RFM12_SDO))) {
      ;
    }


    ///send data bytes
    for (i = 0; i < number_bytes; i++) {

#if RFM12_TRANSMISSION_BURST == 0
      //end of transmission --> Chip Select to '1'
      RFM12_PORT = RFM12_PORT | (1 << RFM12_nSEL);

      _delay_us(0.01);

      //Start of transmission --> Chip Select to '0'
      RFM12_PORT = RFM12_PORT & ~(1 << RFM12_nSEL);
#endif

      SPDR = (tx_data[i] >> 8) & 0xff;         //send high byte to slave

      //wait until data transmission to RFM12 is complete...
      while (!(SPSR & (1 << SPIF))) {
        ;
      }
      SPDR = tx_data[i] & 0xff;                //send low byte to slave

      //wait until data transmission to RFM12 is complete
      while (!(SPSR & (1 << SPIF))) {
        ;
      }

      //wait until RFM12 is ready for the next byte
      while (!(RFM12_PIN & (1 << RFM12_SDO))) {
        ;
      }

    }

    //end of transmission --> Chip Select to '1'
    RFM12_PORT = RFM12_PORT | (1 << RFM12_nSEL);

    _delay_us(0.01);

    //Start of transmission --> Chip Select to '0'
    RFM12_PORT = RFM12_PORT & ~(1 << RFM12_nSEL);


    ///disable TX
    SPDR = (RFM12_POWERMANAGEMENT_RX_VALUE >> 8) & 0xff;         //send high byte to slave

    //wait until data transmission to RFM12 is complete...
    while (!(SPSR & (1 << SPIF))) {
      ;
    }
    *rx_data = SPDR;     //read high byte from slave

    SPDR = RFM12_POWERMANAGEMENT_RX_VALUE & 0xff;         //send low byte to slave

    //wait until data transmission to RFM12 is complete
    while (!(SPSR & (1 << SPIF))) {
      ;
    }

    //end of transmission --> Chip Select to '1'
    RFM12_PORT = RFM12_PORT | (1 << RFM12_nSEL);

    //TX LED off
    RFM12_PORT = RFM12_PORT & ~(1 << RFM12_LEDY);



    ///read RX FIFO
  } else if (di_flag == RFM12_RX_DATA) {


    //RX LED on
    RFM12_PORT = RFM12_PORT | (1 << RFM12_LEDG);

    //Start of transmission --> Chip Select to '0'
    RFM12_PORT = RFM12_PORT & ~(1 << RFM12_nSEL);

    //wait until RFM12 is ready for the next byte
      while (!(RFM12_PIN & (1 << RFM12_SDO))) {
        ;
      }


    ///read data byte

    SPDR = (*tx_data >> 8) & 0xff;         //send high byte to slave

    //wait until data transmission to RFM12 is complete...
    while (!(SPSR & (1 << SPIF))) {
      ;
    }
    *rx_data = SPDR;     //read high byte from slave

    SPDR = *tx_data & 0xff;         //send low byte to slave

    //wait until data transmission to RFM12 is complete
    while (!(SPSR & (1 << SPIF))) {
      ;
    }
    *rx_data = (*rx_data << 8) | SPDR;    //read low byte from slave

    byte_count++;

    //end of transmission --> Chip Select to '1'
    RFM12_PORT = RFM12_PORT | (1 << RFM12_nSEL);

    //TX LED off
    RFM12_PORT = RFM12_PORT & ~(1 << RFM12_LEDG);

  }

  return byte_count;

}


//!///////////////////////////////////////////////////////////////////////
///rfm12_init                                                           //
///                                                                     //
/// Description:                                                        //
///         Initializes the RFM12 Wireless Module after Reset.          //
///                                                                     //
/// input:                                                              //
///         --                                                          //
/// return:                                                             //
///      rfm12_status:   status byte from RFM12 (after initialization)  //
///                                                                     //
//!///////////////////////////////////////////////////////////////////////

uint16_t rfm12_init(void) {

  uint16_t rx_data, tx_data;
#if GLCD_ENABLED == 1
  char tmp_char[16];
#endif

  ///reset RFM12 Wireless Module
  _delay_ms(RFM12_POR_DELAY_MS);

  RFM12_PORT = RFM12_PORT | (1 << RFM12_LEDG) | (1 << RFM12_LEDY);   //reset indication
  RFM12_PORT = RFM12_PORT & ~(1 << RFM12_nRES);
  _delay_ms(RFM12_RESET_WIDTH_MS);
  RFM12_PORT = RFM12_PORT | (1 << RFM12_nRES);
  RFM12_PORT = RFM12_PORT & ~(1 << RFM12_LEDG) & ~(1 << RFM12_LEDY);   //reset indication

  _delay_ms(RFM12_POR_DELAY_MS);


  ///read status register
  tx_data = RFM12_STATUSREAD_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);
#if GLCD_ENABLED == 1
  itoa(rx_data, tmp_char, 16);
  glcd_send_string("\nRFM12 Status: ", FALSE);
  glcd_send_string(tmp_char, FALSE);
  _delay_ms(RFM12_POR_DELAY_MS);
  _delay_ms(RFM12_POR_DELAY_MS);
#endif

  ///send RFM12 Config Command
  tx_data = RFM12_CONFIG_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);

  ///send RFM12 Power Management Command
  tx_data = RFM12_POWERMANAGEMENT_RX_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);

  ///send RFM12 Frequency Command
  tx_data = RFM12_FREQUENCY_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);

  ///send RFM12 Data Rate Command
  tx_data = RFM12_DATARATE_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);

  ///send RFM12 Power Setting Command
  tx_data = RFM12_POWERSETTING_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);

  ///send RFM12 Data Filter Command
  tx_data = RFM12_DATAFILTER_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);

  ///send RFM12 FIFO Mode Command
  tx_data = RFM12_FIFOMODE_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);

  ///send RFM12 AFC Command
  tx_data = RFM12_AFC_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);

  ///send RFM12 TX Config Command
  tx_data = RFM12_TXCONFIG_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);

  ///send RFM12 PLL Setting Command
  tx_data = RFM12_PLLSETTING_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);

  ///send RFM12 Wakeup Timer Command
  tx_data = RFM12_WAKEUPTIMER_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);

  ///send RFM12 Low Duty-Cycle Command
  tx_data = RFM12_LOWDUTYCYCLE_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);

  ///send RFM12 Synchron Pattern Command
  tx_data = RFM12_SYNCHRONPATTERN_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);

  ///send RFM12 Low Battery Detector Command
  tx_data = RFM12_LOWBATDET_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);


  ///read status register
  tx_data = RFM12_STATUSREAD_VALUE;
  rfm12_data_io(&tx_data, &rx_data, 1, RFM12_REGISTER_DATA);
#if GLCD_ENABLED == 1
  itoa(rx_data, tmp_char, 16);
  glcd_send_string("\nRFM12 Status: ", FALSE);
  glcd_send_string(tmp_char, FALSE);
  _delay_ms(RFM12_POR_DELAY_MS);
  _delay_ms(RFM12_POR_DELAY_MS);
  _delay_ms(RFM12_POR_DELAY_MS);
  _delay_ms(RFM12_POR_DELAY_MS);
  _delay_ms(RFM12_POR_DELAY_MS);
#endif

  _delay_ms(RFM12_POR_DELAY_MS);

  return rx_data;


}




//!///////////////////////////////////////////////////////////////////////
///rfm12_pin_init                                                       //
///                                                                     //
/// Description:                                                        //
///         Initializes the pins for interfacing the RFM12 Module.      //
///                                                                     //
/// input:                                                              //
///         --                                                          //
/// return:                                                             //
///         --                                                          //
///                                                                     //
//!///////////////////////////////////////////////////////////////////////

void rfm12_pin_init(void) {

  //setup port (values)
  RFM12_PORT = RFM12_PORT | (1 << RFM12_nIRQ) | (1 << RFM12_nRES) | (1 << RFM12_nSEL) |
               (1 << RFM12_SDI) | (1 << RFM12_SDO) | (1 << RFM12_SCK);       //RFM12_LEDG and RFM12_LEDY = 0x00
  RFM12_PORT = RFM12_PORT & ~(1 << RFM12_LEDG) & ~(1 << RFM12_LEDY);


  //set data direction register
  RFM12_DDR = RFM12_DDR | (1 << RFM12_LEDG) | (1 << RFM12_LEDY) | (1 << RFM12_nRES)
              | (1 << RFM12_SDI) | (1 << RFM12_SCK) | (1 << RFM12_nSEL);                  //output
  RFM12_DDR = RFM12_DDR & ~(1 << RFM12_nIRQ) & ~(1 << RFM12_SDO);                         //input


  //enable SPI interface
  SPCR = 0x00 | (1 << SPE) | (1 << MSTR) | (1 << SPR0);    //SPI enabled, Master, MCU_CLK/16


}


