//!/////////////////////////////////////////////////////////////////////////
/// usart.c                                                               //
///                                                                       //
/// Version:     0.3                                                      //
/// Date:        05/22/2010                                               //
/// Author:      Michael Skutek (michael@skutek.de)                       //
///                                                                       //
/// Description: Contains the USART routines                              //
///                                                                       //
/// Platform:    AVR ATmega (tested on ATmega16)                          //
///                                                                       //
///                                                                       //
///                                                                       //
/// This program is free software: you can redistribute it and/or modify  //
/// it under the terms of the GNU Lesser General Public License as        //
/// published by the Free Software Foundation, either version 3 of the    //
/// License, or (at your option) any later version.                       //
///                                                                       //
/// This program is distributed in the hope that it will be useful,       //
/// but WITHOUT ANY WARRANTY; without even the implied warranty of        //
/// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the          //
/// GNU Lesser General Public License for more details.                   //
///                                                                       //
/// You should have received a copy of the GNU Lesser General Public      //
/// License along with this program. If not, see                          //
/// <http://www.gnu.org/licenses/>.                                       //
//!/////////////////////////////////////////////////////////////////////////


#include <avr/io.h>
#include <stdint.h>
#include <util/delay.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>

#include "babyfone_receiver.h"
#include "port_definitions.h"
#include "usart.h"


///global variables for ISR
extern volatile char usart_rcv_buffer[USART_RCV_BUFFER_LENGTH];
extern volatile uint8_t usart_rcv_position;

extern volatile uint8_t print_command_error;
extern volatile uint8_t print_command_info;


//!/////////////////////////////////////////////
///Init Function for USART//////////////////////
//!/////////////////////////////////////////////

void USART_Init(unsigned int baud) {

  //set baud rate
#if MCU == 644

  UBRR0H = (unsigned char)(baud>>8);
  UBRR0L = (unsigned char)baud;

  UCSR0B = (1 << RXEN0) | (1 << TXEN0) | (1 << RXCIE0);               //activate RX, TX and RX Complete Interrupt
  UCSR0C = (1 << UCSZ01) | (1 << UCSZ00);                             //asynchronouos 8N1

#else

  UBRRH = (unsigned char)(baud>>8);
  UBRRL = (unsigned char)baud;

  UCSRB = (1 << RXEN) | (1 << TXEN) | (1 << RXCIE);               //activate RX, TX and RX Complete Interrupt
  UCSRC = (1 << URSEL) | (1 << UCSZ1) | (1 << UCSZ0);             //asynchronouos 8N1

#endif



}


//!/////////////////////////////////////////////
///Send single character via USART//////////////
//!/////////////////////////////////////////////

void USART_putc(unsigned char c) {

#if MCU == 644

  while (!(UCSR0A & (1<<UDRE0))) { //wait until sendig is possible

  }

  UDR0 = c;                     //send single char

#else

  while (!(UCSRA & (1<<UDRE))) { //wait until sendig is possible

  }

  UDR = c;                      //send single char

#endif

}


//!/////////////////////////////////////////////
///Send string via USART (using USART_putc)/////
//!/////////////////////////////////////////////

void USART_puts(const char *s, uint8_t flash_mem) {

  char c;

  if (flash_mem == TRUE) {
    c = pgm_read_byte(s);
    while (c != '\0') {   //until *s != '\0' --> means "string end sign"
      USART_putc(c);
      s++;
      c = pgm_read_byte(s);
    }

  } else {

    while (*s != '\0') {   //until *s != '\0' --> means "string end sign"
      USART_putc(*s);
      s++;
    }

  }
}


//!/////////////////////////////////////////////
///Process USART commands///////////////////////
//!/////////////////////////////////////////////

uint8_t process_usart_command(void) {

  uint8_t i;
  uint16_t checksum;


  //check for lenght of string to avoid overflow errors
  if (usart_rcv_position < 3) {
    return USART_UNKNOWN_COMMAND;
  }


  //calculate checksum of received command
  checksum = 0;
  for (i = 0; i < usart_rcv_position; i++) {
    checksum = checksum + (uint8_t) usart_rcv_buffer[i];
  }


  if (checksum == SYNCHRONIZE) {

    //print_command_info = UART_COUNTER_MSG;

    return USART_COMMAND_OK;

  }

  return USART_UNKNOWN_COMMAND;

}
